-----------------------------------------------------------------------
-- To do and/or ideas
--
-- From Rabbit:
-- * Parse damage messages from swstats and damagemeters.
-- * Automatically output our damage done and position to chat when a boss has
-- been killed; this is easy with Babble-Boss, CHAT_MSG_HOSTILE_DEATH and
-- UnitClassification.
-- * Maybe add an auto-reset for boss kill as well.
--
-- From Nymbia:
-- * Pick up the slack of users without the addon by splitting up the members
-- without the addon among those with it.  ie, in a 25 man raid where 10 users
-- have Violation, each of them is assigned to parse the damage of either 1 or
-- 2 other raid members.
-- * Possibly keep two running totals:  'this fight' and 'overall', to allow
-- for statistics to be displayed for a single encounter without requiring
-- constant resets.
--
-- From <person>:
-- * Brew my coffee when I win the damage meter contest after a boss fight.
--
-----------------------------------------------------------------------

-----------------------------------------------------------------------
-- Locale
-----------------------------------------------------------------------
local L = AceLibrary("AceLocale-2.2"):new("Violation")

L:RegisterTranslations("enUS", function() return {
	["Track"] = true,
	["What data types to track."] = true,
	["Tooltip"] = true,
	["What data to show in the tooltip."] = true,
	["Toggle whether to display data from %q in the tooltip."] = true,
	["Colors"] = true,
	["Assign colors to the different data types."] = true,
	["What color to use for the %q data type."] = true,
	["Color the tooltip"] = true,
	["Toggles whether or not to use these colors in the tooltip."] = true,

	["Resetting data as requested by %s."] = true,
	["Reset"] = true,
	["Options for when to reset the data."] = true,
	["Reset now"] = true,
	["Immediately resets your data."] = true,
	["Reset group"] = true,
	["Resets the Violation data for every group member. Only the group leader or an officer can do this."] = true,
	["Leader resets"] = true,
	["Accept reset requests from raid leaders or officers."] = true,
	["Auto-reset"] = true,
	["Entering combat"] = true,
	["Automatically reset the data when entering combat."] = true,
	["Leaving combat"] = true,
	["Automatically reset the data when leaving combat."] = true,
	["Print on reset"] = true,
	["Prints your current data locally when your data is reset."] = true,

	["Text"] = true,
	["What data to show on the FuBar panel, if available."] = true,

	["Toggle whether to track %q."] = true,
	["Toggle whether to display data from %q on the FuBar panel."] = true,

	["%s has been disabled, since it is not compatible with the latest module API."] = true,
	["No modules running, please enable one or more from the configuration menu."] = true,
	["No data gathered yet."] = true,
	["|cffeda55fClick|r to reset."] = true,
	["Player"] = true,
	["#"] = true,
	["%s: %d"] = true,
} end)

L:RegisterTranslations("koKR", function() return {
	["Track"] = "측정",
	["What data types to track."] = "측정할 데이터를 선택합니다.",
	["Tooltip"] = "툴팁",
	["What data to show in the tooltip."] = "툴팁에 표시할 데이터를 선택합니다.",
	["Toggle whether to display data from %q in the tooltip."] = "툴팁에 %q 데이터 표시를 전환합니다.",
	["Colors"] = "색상",
	["Assign colors to the different data types."] = "각각의 데이터 종류에 색상을 지정합니다.",
	["What color to use for the %q data type."] = "%q 데이터에 사용할 색상을 지정합니다.",
	["Color the tooltip"] = "툴팁 색상",
	["Toggles whether or not to use these colors in the tooltip."] = "툴팁 색상 사용을 전환합니다.",

	["Resetting data as requested by %s."] = "%s의 요청에 의해 데이터를 초기화합니다.",
	["Reset"] = "초기화",
	["Options for when to reset the data."] = "데이터 초기화에 대한 설정입니다.",
	["Reset now"] = "즉시 초기화",
	["Immediately resets your data."] = "당신의 데이터를 즉시 초기화합니다.",
	["Reset group"] = "파티 초기화",
	["Resets the Violation data for every group member. Only the group leader or an officer can do this."] = "모든 파티원에 대한 Violation 데이터를 초기화합니다. 단, 공대장 혹은 승급자 이어야 합니다.",
	["Leader resets"] = "리더 초기화",
	["Accept reset requests from raid leaders or officers."] = "공대장 혹은 오피서로 부터의 초기화 요청을 허용합니다.",
	["Auto-reset"] = "자동 초기화",
	["Entering combat"] = "전투 시작시",
	["Automatically reset the data when entering combat."] = "전투 시작시 자동으로 데이터를 초기화합니다.",
	["Leaving combat"] = "전투 종료시",
	["Automatically reset the data when leaving combat."] = "전투 종료시 자동으로 데이터를 초기화합니다.",
	["Print on reset"] = "초기화 시 출력",
	["Prints your current data locally when your data is reset."] = "당신의 데이터를 초기화할 때 현재 데이터를 출력합니다.",

	["Text"] = "텍스트",
	["What data to show on the FuBar panel, if available."] = "FuBar에 표시할 데이터를 선택합니다.",

	["Toggle whether to track %q."] = "%q 기록 전환",
	["Toggle whether to display data from %q on the FuBar panel."] = "FuBar에 %q 데이터 표시를 전환합니다.",

	["No modules running, please enable one or more from the configuration menu."] = "실행중인 모듈이 없습니다. 환경설정 메뉴에서 하나 이상을 활성화하세요.",
	["No data gathered yet."] = "집계된 데이터가 없습니다.",
	["|cffeda55fClick|r to reset."] = "초기화 하려면 |cffeda55f클릭|r하세요.",
} end)

-----------------------------------------------------------------------
-- Locals
-----------------------------------------------------------------------
local tablet = AceLibrary("Tablet-2.0")
local BC = AceLibrary("Babble-Class-2.2")
local roster = AceLibrary("Roster-2.1")

local coloredNames = setmetatable({}, {__index =
	function(self, key)
		local obj = roster:GetUnitObjectFromName(key)
		if not obj then return nil end
		self[key] = "|cff" .. BC:GetHexColor(obj.class) .. key .. "|r"
		return self[key]
	end
})

local db = nil

local _G = getfenv(0)
local table_insert = table.insert

local playerName = nil
local percentString = "(%d%%)"

local moduleVersion = 3

local runningModules = {}
local violationUsers = nil

local red = "Interface\\AddOns\\Violation\\icons\\red"
local blue = "Interface\\AddOns\\Violation\\icons\\blue"
local grey = "Interface\\AddOns\\Violation\\icons\\grey"

-----------------------------------------------------------------------
-- Local heap and utility functions
-----------------------------------------------------------------------

local new, del
do
	local cache = setmetatable({},{__mode='k'})
	function new(...)
		local t = next(cache)
		if t then
			cache[t] = nil
			for i = 1, select("#", ...) do
				table_insert(t, (select(i, ...)))
			end
			return t
		else
			return {...}
		end
	end
	function del(t)
		if type(t) == "table" then
			for k, v in pairs(t) do
				if type(v) == "table" then
					del(v)
				end
				t[k] = nil
			end
			cache[t] = true
		end
		return nil
	end
end

local function Colorize(text, r, g, b)
	return string.format("|cff%02x%02x%02x%s|r", r * 255, g * 255, b * 255, text)
end

-----------------------------------------------------------------------
-- Options
-----------------------------------------------------------------------

local options = {
	type = "group",
	args = {
		track = {
			type = "group",
			name = L["Track"],
			desc = L["What data types to track."],
			disabled = function() return not Violation:IsActive() end,
			args = {}, -- Auto-populated by module toggles.
			order = 101,
		},
		fubar = {
			type = "group",
			name = L["Text"],
			desc = L["What data to show on the FuBar panel, if available."],
			disabled = function() return not FuBar end,
			args = {}, -- Auto-populated by module toggles.
			order = 102,
		},
		colors = {
			type = "group",
			name = L["Colors"],
			desc = L["Assign colors to the different data types."],
			args = {
				tooltip = {
					type = "toggle",
					name = L["Color the tooltip"],
					desc = L["Toggles whether or not to use these colors in the tooltip."],
					get = function() return Violation.db.char.colorTooltip end,
					set = function(v)
						Violation.db.char.colorTooltip = v
						Violation:UpdateTooltip()
					end,
					order = 1,
					disabled = true,
				},
				spacer = {
					type = "header",
					name = " ",
					order = 50,
				},
			}, -- Auto-populated by module toggles.
			order = 103,
		},
		tooltip = {
			type = "group",
			name = L["Tooltip"],
			desc = L["What data to show in the tooltip."],
			args = {},
			order = 104,
		},
		spacer = {
			type = "header",
			name = " ",
			order = 150,
		},
		print = {
			type = "toggle",
			name = L["Print on reset"],
			desc = L["Prints your current data locally when your data is reset."],
			get = function() return Violation.db.profile.printOnReset end,
			set = function(v) Violation.db.profile.printOnReset = v end,
			order = 200,
		},
		reset = {
			type = "group",
			name = L["Reset"],
			desc = L["Options for when to reset the data."],
			args = {
				now = {
					type = "execute",
					name = L["Reset now"],
					desc = L["Immediately resets your data."],
					func = function() Violation:Reset() end,
					order = 100,
				},
				group = {
					type = "execute",
					name = L["Reset group"],
					desc = L["Resets the Violation data for every group member. Only the group leader or an officer can do this."],
					func = function() Violation:ResetGroup() end,
					disabled = function()
						return not IsRaidLeader() and not IsRaidOfficer() and not IsPartyLeader()
					end,
					order = 101,
				},
				leader = {
					type = "toggle",
					name = L["Leader resets"],
					desc = L["Accept reset requests from raid leaders or officers."],
					get = function() return Violation.db.profile.leaderReset end,
					set = function(v) Violation.db.profile.leaderReset = v end,
					order = 102,
				},
				spacer = {
					type = "header",
					name = " ",
					order = 150,
				},
				autoheader = {
					type = "header",
					name = L["Auto-reset"],
					order = 200,
				},
				gaincombat = {
					type = "toggle",
					name = L["Entering combat"],
					desc = L["Automatically reset the data when entering combat."],
					get = function() return Violation:IsEventRegistered("PLAYER_REGEN_DISABLED") end,
					set = function(v)
						if v then
							Violation:RegisterEvent("PLAYER_REGEN_DISABLED", "Reset")
						else
							Violation:UnregisterEvent("PLAYER_REGEN_DISABLED")
						end
						Violation.db.profile.gainCombat = v
					end,
					order = 201,
				},
				leavecombat = {
					type = "toggle",
					name = L["Leaving combat"],
					desc = L["Automatically reset the data when leaving combat."],
					get = function() return Violation:IsEventRegistered("PLAYER_REGEN_ENABLED") end,
					set = function(v)
						if v then
							Violation:RegisterEvent("PLAYER_REGEN_ENABLED", "Reset")
						else
							Violation:UnregisterEvent("PLAYER_REGEN_ENABLED")
						end
						Violation.db.profile.leaveCombat = v
					end,
					order = 202,
				},
			},
			order = 201,
		},
	},
}

-----------------------------------------------------------------------
-- Addon declaration
-----------------------------------------------------------------------
Violation = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceDB-2.0", "AceConsole-2.0", "AceComm-2.0", "FuBarPlugin-2.0", "AceModuleCore-2.0", "Parser-3.0")
local Violation = Violation

Violation:SetModuleMixins("AceEvent-2.0")
Violation.hasIcon = true

-----------------------------------------------------------------------
-- Module prototype
-----------------------------------------------------------------------

Violation.modulePrototype.core = Violation
Violation.modulePrototype.version = -1

local incompatible = {}
local registered = {}
function Violation.modulePrototype:OnInitialize()
	if type(self.name) ~= "string" then
		error("A module is missing its name field.")
	elseif type(self.version) ~= "number" then
		error(self.name .. " doesn't have a version field.")
	elseif registered[self.name] then
		error("A module has been initialized twice for some reason.")
	end

	if self.version == moduleVersion then
		self.core:OnRegister(self)
		registered[self.name] = true
	else
		incompatible[self] = true
		self.core:ModuleIncompatible(self)
	end
end

function Violation.modulePrototype:OnEnable(first)
	if incompatible[self] then return end

	if not runningModules[self.name] then
		-- Always reset our data when we enable.
		self:OnReset()

		if type(self.OnModuleEnable) == "function" then
			self:OnModuleEnable()
		end

		runningModules[self.name] = self

		-- self.core:SendCommMessage("GROUP", "ModuleEnabled", self.name)
	end
end

function Violation.modulePrototype:OnDisable()
	if runningModules[self.name] then
		if type(self.OnModuleDisable) == "function" then
			self:OnModuleDisable()
		end

		runningModules[self.name] = nil

		-- self.core:SendCommMessage("GROUP", "ModuleDisabled", self.name)
	end
end

function Violation.modulePrototype:new(...) return new(...) end
function Violation.modulePrototype:del(t) return del(t) end

-----------------------------------------------------------------------
-- Implicit module interface
-----------------------------------------------------------------------

local AceOO = AceLibrary("AceOO-2.0")
local IViolation = AceOO.Interface{

	-- Optional methods:
	--
	-- -- Called when the module is enabled by the core. Any event registration
	-- -- should happen here.
	-- OnModuleEnable
	-- OnModuleDisable
	--
	-- -- Called when the module is initially registered.
	-- OnModuleRegister

	-- Must return the current module revision as a number.
	GetRevision = "function",

	-- Must return the module name as it should be displayed in the menu, for
	-- example "Damage done".
	GetDisplayName = "function",

	-- Must return the module name as it should be displayed in a condensed
	-- display, like "PlayerDPS" should return "DPS".
	GetShortDisplayName = "function",

	-- Must return whether the modules data is dirty or not. If it is, a
	-- broadcast will be scheduled by the core.
	IsDirty = "function",

	-- Must return the data that should be broadcasted to the other players.
	-- Must return 1 number only.
	GetBroadcastData = "function",

	-- Invoked when remote data is received for this module. Usually invoked by
	-- the core, but can also be invoked by other modules when they parse data
	-- from other addons, for example.
	-- Will only get two arguments, the first will be the person the data
	-- belongs to and the second will be a number.
	OnRemoteData = "function",

	-- Invoked by the core when all data should be reset.
	OnReset = "function",

	-- Must return a universal unique ID that will be used for this module when
	-- sending data (where the universe is Violation)
	-- (Not used now, we use a memoization of the module name instead.
	-- GetUUID = "function",

	-- Must return the current dataset for this module. If an argument is given,
	-- it can be assumed to be a player name, and only the data (a number) for
	-- that player (or nil) must be returned.
	GetData = "function",
}

-----------------------------------------------------------------------
-- Addon initialization
-----------------------------------------------------------------------

function Violation:OnInitialize()
	local revision = tonumber(("$Revision: 29303 $"):sub(12, -3)) or 1
	if not self.version then self.version = "1.0" end
	self.version = self.version .. "." .. revision
	self.revision = revision

	_G.VIOLATION_VERSION = self.version
	_G.VIOLATION_REVISION = self.revision

	self:RegisterDB("ViolationDB", "ViolationPerCharDB", "char")
	self:RegisterDefaults("char", {
		enabledModules = {},
		textModules = {},
		tooltipModules = {},
		colors = {
			["*"] = {
				r = 1, g = 1, b = 1,
			}
		},
		colorTooltip = true,
		gainCombat = nil,
		leaveCombat = nil,
		leaderReset = true,
		printOnReset = true,
	})

	self.OnMenuRequest = options

	self:RegisterChatCommand({ "/violation" }, self.OnMenuRequest, "VIOLATION")

	self:SetCommPrefix("Violation")
	self:SetDefaultCommPriority("BULK")

	self:SetIcon(grey)

	playerName = UnitName("player")
end

function Violation:OnEnable(first)
	db = self.db.char

	if first then
		local memoizations = new()
		for name in self:IterateModules() do
			table_insert(memoizations, name)
		end
		table_insert(memoizations, "LeaderReset")
		table_insert(memoizations, "ViolationPing")
		table_insert(memoizations, "ViolationPong")
		self:RegisterMemoizations(memoizations)
		memoizations = del(memoizations)
	end

	-- Enable/disable all modules.
	for name, mod in self:IterateModules() do
		self:ToggleModuleActive(name, db.enabledModules[name])
	end

	self:Reset()

	self:RegisterComm("Violation", "GROUP", "OnComm")
	self:RegisterComm("Violation", "WHISPER", "OnPrivateComm")

	self:ScheduleRepeatingEvent("Broadcast", self.Broadcast, 5, self)
	self:ScheduleRepeatingEvent("DisplayUpdate", self.UpdateDisplayIfDirty, 2, self)

	if db.gainCombat then
		self:RegisterEvent("PLAYER_REGEN_DISABLED", "Reset")
	end

	if db.leaveCombat then
		self:RegisterEvent("PLAYER_REGEN_ENABLED", "Reset")
	end

	self:RegisterBucketEvent({"RAID_ROSTER_UPDATE", "PARTY_MEMBERS_CHANGED"}, 2, function()
		violationUsers = del(violationUsers)
		self:SendCommMessage("GROUP", "ViolationPing")
	end)

	self:SetIcon(red)
end

function Violation:OnDisable()
	-- Disable all modules.
	for name, mod in pairs(runningModules) do
		self:ToggleModuleActive(name, false)
	end

	self:SetIcon(grey)
end
function Violation:OnTextUpdate()
	local hasText = nil
	for k, v in pairs(db.textModules) do
		if v and runningModules[k] then
			local data = runningModules[k]:GetData(playerName)
			if type(data) == "number" and data > 0 then
				local shortName = runningModules[k]:GetShortDisplayName()
				local color = db.colors[k]
				local text = Colorize(string.format(L["%s: %d"], shortName, data), color.r, color.g, color.b)
				if not hasText then hasText = new(text)
				else table_insert(hasText, text) end
			end
		end
	end

	if not hasText then
		self:SetText(self.name)
	else
		self:SetText(table.concat(hasText, " "))
		hasText = del(hasText)
	end
end


function Violation:Reset()
	for name, mod in pairs(runningModules) do
		if db.printOnReset then
			local data = mod:GetData(playerName)
			if type(data) == "number" and data > 0 then
				local name = mod:GetDisplayName()
				local color = db.colors[name]
				self:Print(Colorize(string.format(L["%s: %d"], name, data), color.r, color.g, color.b))
			end
		end
		mod:OnReset()
	end
	self:UpdateDisplay()
end

-----------------------------------------------------------------------
-- Module handling
-----------------------------------------------------------------------

function Violation:OnRegister(mod)
	if not AceOO.inherits(mod, IViolation) then
		error(string.format("The module %q does not inherit from the IViolation interface.", tostring(mod.name)))
	end

	local moduleName = mod.name
	local displayName = mod:GetDisplayName()

	options.args.track.args[moduleName] = {
		type = "toggle",
		name = displayName,
		desc = string.format(L["Toggle whether to track %q."], displayName),
		get = function()
			return self.db.char.enabledModules[moduleName]
		end,
		set = function(v)
			self.db.char.enabledModules[moduleName] = v
			self:ToggleModuleActive(moduleName, v)
			self:UpdateDisplay()
		end
	}

	options.args.fubar.args[moduleName] = {
		type = "toggle",
		name = displayName,
		desc = string.format(L["Toggle whether to display data from %q on the FuBar panel."], displayName),
		get = function()
			return self.db.char.textModules[moduleName]
		end,
		set = function(v)
			self.db.char.textModules[moduleName] = v
			self:UpdateText()
		end
	}

	options.args.tooltip.args[moduleName] = {
		type = "toggle",
		name = displayName,
		desc = string.format(L["Toggle whether to display data from %q in the tooltip."], displayName),
		get = function()
			return self.db.char.tooltipModules[moduleName]
		end,
		set = function(v)
			self.db.char.tooltipModules[moduleName] = v
			self:UpdateTooltip()
		end
	}

	options.args.colors.args[moduleName] = {
		type = "color",
		name = displayName,
		desc = string.format(L["What color to use for the %q data type."], displayName),
		get = function()
			local tuple = self.db.char.colors[moduleName]
			return tuple.r, tuple.g, tuple.b
		end,
		set = function(r, g, b)
			local tuple = self.db.char.colors[moduleName]
			tuple.r = r
			tuple.g = g
			tuple.b = b
			self:ScheduleEvent("ColorUpdate", self.UpdateDisplay, 5, self)
		end,
		order = 100,
	}

	if type(mod.OnModuleRegister) == "function" then
		mod:OnModuleRegister()
	end
end

function Violation:GetModuleVersion()
	return moduleVersion
end

function Violation:ModuleIncompatible(mod)
	self:ToggleModuleActive(mod, false)
	self:Print(string.format(L["%s has been disabled, since it is not compatible with the latest module API."], mod.name))
end

-----------------------------------------------------------------------
-- UI
-----------------------------------------------------------------------

function Violation:OnClick()
	self:Reset()
end

function Violation:UpdateDisplayIfDirty()
	local dirty = nil
	for k, v in pairs(db.tooltipModules) do
		if v and runningModules[k] and runningModules[k]:IsDirty() then
			dirty = true
		end
	end
	if dirty then
		self:UpdateDisplay()
	end
end

local function Sorter(a, b) return a[2] > b[2] end

local function AddModuleData(name, data)
	if not data then return end

	local tbl = new()
	for k, v in pairs(data) do
		if type(v) == "number" and v > 0 then
			table_insert(tbl, new(k, v))
		end
	end

	local tblCount = #tbl
	if tblCount > 0 then
		local cat = tablet:AddCategory(
			"columns", tblCount > 1 and 4 or 2,
			"text", tblCount > 1 and L["#"] or name,
			"text2", tblCount > 1 and L["Player"] or "",
			"text3", name,
			"child_justify", "LEFT",
			"child_justify2", tblCount > 1 and "LEFT" or "RIGHT",
			"child_justify3", "RIGHT",
			"child_justify4", "RIGHT",
			"child_textR", 1,
			"child_textG", 1,
			"child_textB", 0,
			"child_text2R", 1,
			"child_text2G", 1,
			"child_text2B", 1,
			"child_text3R", 1,
			"child_text3G", 1,
			"child_text3B", 1,
			"child_text4R", 0.65,
			"child_text4G", 0.65,
			"child_text4B", 0.65
		)

		local percentage = nil
		if tblCount > 1 then
			table.sort(tbl, Sorter)

			local total = 0
			for i, v in ipairs(tbl) do
				total = total + v[2]
			end
			percentage = new()
			for i, v in ipairs(tbl) do
				table_insert(percentage, math.floor((v[2] * 100) / total))
			end
		end

		local playerShown = nil
		for i, v in ipairs(tbl) do
			if i > 10 then break end -- Only show the top 10.
			if percentage then
				cat:AddLine(
					"text", tostring(i) .. ".",
					"text2", coloredNames[v[1]],
					"text3", tostring(v[2]),
					"text4", string.format(percentString, percentage[i])
				)
			else
				cat:AddLine(
					"text", coloredNames[v[1]],
					"text2", tostring(v[2])
				)
			end
			if v[1] == playerName then
				playerShown = true
			end
		end

		percentage = del(percentage)

		if not playerShown and data[playerName] > 0 then
			cat:AddLine("text", "?", "text2", coloredNames[playerName], "text3", tostring(data[playerName]))
		end
	end

	-- Cleanup.
	tbl = del(tbl)
end


function Violation:OnTooltipUpdate()
	local no = true
	for k, v in pairs(db.tooltipModules) do
		if v and runningModules[k] then
			no = nil
			local data = runningModules[k]:GetData()
			assert(type(data) == "table", "Some module provided us with nil data.")
			local name = runningModules[k]:GetDisplayName()
			AddModuleData(name, data)
		end
	end
	if no then
		tablet:AddCategory("columns", 1):AddLine("text", L["No modules running, please enable one or more from the configuration menu."])
	end
	tablet:SetHint(L["|cffeda55fClick|r to reset."])
end

function Violation:OnTextUpdate()
	local hasText = nil
	for k, v in pairs(db.textModules) do
		if v and runningModules[k] then
			local data = runningModules[k]:GetData(playerName)
			if type(data) == "number" and data > 0 then
				local shortName = runningModules[k]:GetShortDisplayName()
				local color = db.colors[k]
				local text = Colorize(string.format(L["%s: %d"], shortName, data), color.r, color.g, color.b)
				if not hasText then hasText = new(text)
				else table_insert(hasText, text) end
			end
		end
	end

	if not hasText then
		self:SetText(self.name)
	else
		self:SetText(table.concat(hasText, " "))
		hasText = del(hasText)
	end
end

-----------------------------------------------------------------------
-- Data broadcasting and recieving
-----------------------------------------------------------------------
function Violation:Broadcast()
	local tbl = nil
	for name, mod in pairs(runningModules) do
		if mod:IsDirty() then
			if not tbl then tbl = new() end
			table_insert(tbl, name)
			table_insert(tbl, mod:GetBroadcastData())
		end
	end
	if tbl then
		self:SendCommMessage("GROUP", "B", unpack(tbl))
	end
	tbl = del(tbl)
end

function Violation:ResetGroup()
	if not IsRaidLeader() and not IsRaidOfficer() and not IsPartyLeader() then
		error("Only the group leader or an officer can do that.")
		return
	end

	self:SendCommMessage("GROUP", "LeaderReset")
	self:Reset()
end

function Violation:OnPrivateComm(prefix, sender, distribution, command, ...)
	if distribution ~= "WHISPER" or type(command) ~= "string" then return end
	if command == "ViolationPong" then
		if type(violationUsers) ~= "table" then
			violationUsers = new()
		end
		violationUsers[sender] = new(...)
	end
end

function Violation:OnComm(prefix, sender, distribution, command, ...)
	if distribution ~= "GROUP" or type(command) ~= "string" then return end

	if command == "B" then
		local total = select("#", ...)
		local i
		for i = 1, total do
			local name = select(i, ...)
			if runningModules[name] then
				runningModules[name]:OnRemoteData(sender, select(i + 1, ...))
			end
			i = i + 1
			if i > total then break end
		end
	elseif command == "LeaderReset" and db.leaderReset then
		local obj = roster:GetUnitObjectFromName(sender)
		if obj.rank > 0 then
			self:Print(string.format(L["Resetting data as requested by %s."], sender))
			self:Reset(sender)
		end
	elseif command == "ViolationPing" then
		local tbl = new()
		for k in pairs(runningModules) do
			table_insert(tbl, k)
		end
		self:SendCommMessage("WHISPER", sender, "ViolationPong", self.version, self.revision, unpack(tbl))
		tbl = del(tbl)
	end
end

